/*	
 *	d12ci.c
 *
 *	PDIUSBD12 R}hC^[tF[X
 *	Copyright (c) 1997 Philips Semiconductors
 *	Copyright (c) 2002 Naoyuki Sawa
 *
 *	* Mon Aug  5 04:47:00 JST 2002 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "d12ci.h"

/****************************************************************************
 *	萔
 ****************************************************************************/

#define D12_COMMAND	0x400004	/* R}hWX^̃AhX */
#define D12_DATA	0x400000	/* f[^WX^̃AhX   */

/* Initialization commands */
#define SET_ADDRESS_ENABLE		0xd0
#define SET_ENDPOINT_ENABLE		0xd8
#define SET_MODE			0xf3
#define SET_DMA				0xfb
/* Data flow commands */
#define READ_INTERRUPT_REGISTER		0xf4
#define SELECT_ENDPOINT			0x00	/* 0x00`0x05 */
#define READ_LAST_TRANSACTION_STATUS	0x40	/* 0x40`0x45 */
#define READ_BUFFER			0xf0
#define WRITE_BUFFER			0xf0
#define SET_ENDPOINT_STATUS		0x40	/* 0x40`0x45 */
#define ACKNOWLEDGE_SETUP		0xf1
#define CLEAR_BUFFER			0xf2
#define VALIDATE_BUFFER			0xfa
#define SEND_RESUME			0xf6
#define READ_CURRENT_FRAME_NUMBER	0xf5
/* Undocumented commands */
#define READ_CHIP_ID			0xfd

/****************************************************************************
 *	[J֐
 ****************************************************************************/

/* * f[^|[g񖽗ߘAœǂݍނƁAlǂݏo܂B
 *   Ⴆ΁Â悤ȃR[hƁAȂȂ܂B
 *	xld.w %r0, 0x400000
 *	xld.w %r1, [%r0]
 *	xld.w %r2, [%r0]
 *   D12_ReadEndpoint()ł́A_~[ƃTCY̓ǂݍ݂Ȃǂ댯łB
 * * ͕słACPU`USBCւCE7MɊ܂܂ĂA
 *   fBCHł͂ȂƎv܂B
 *   (KvȃfBCȂ̂ł傤ǁÂԂɉeoĂ?)
 * * ۂɂ́Ainportb()֐ĂяõI[o[wbĥŁA
 *   CCWJȂÂ悤ȃR[h͐Ȃ͂łB
 *   AYȂ悤(Ô߂)AInopނƂɂ܂B
 *	xld.w %r0, 0x400000
 *	xld.w %r1, [%r0]
 *	nop I
 *	xld.w %r2, [%r0]
 *	nop I
 * * f[^|[gւ̓Aߏ݂͌؂Ă܂񂪁A
 *   Ԃ񓯂soƎv̂ŁAnopނƂɂ܂B
 */

static void
outportb(unsigned long port, unsigned char val)
{
	*(volatile unsigned char*)port = val;
	asm("nop"); /* KvI */
}

static unsigned char
inportb(unsigned long port)
{
	unsigned char val;
	val = *(volatile unsigned char*)port;
	asm("nop"); /* KvI */
	return val;
}

/****************************************************************************
 *	O[o֐
 ****************************************************************************/

void
D12_SetAddressEnable(unsigned char addr, unsigned char enable)
{
	outportb(D12_COMMAND, SET_ADDRESS_ENABLE);
	if(enable) addr |= 0x80;
	outportb(D12_DATA, addr);
}

void
D12_SetEndpointEnable(unsigned char enable)
{
	outportb(D12_COMMAND, SET_ENDPOINT_ENABLE);
	outportb(D12_DATA, enable ? 0x01 : 0x00);
}

void
D12_SetMode(unsigned char config, unsigned char clkdiv)
{
	outportb(D12_COMMAND, SET_MODE);
	outportb(D12_DATA, config);
	outportb(D12_DATA, clkdiv);
}

void
D12_SetDMA(unsigned char mode)
{
	outportb(D12_COMMAND, SET_DMA);
	outportb(D12_DATA, mode);
}

unsigned short
D12_ReadInterruptRegister()
{
	unsigned char lo, hi;
	outportb(D12_COMMAND, READ_INTERRUPT_REGISTER);
	lo = inportb(D12_DATA);
	hi = inportb(D12_DATA);
	return hi << 8 | lo;
}

unsigned char
D12_SelectEndpoint(unsigned char endp)
{
	outportb(D12_COMMAND, SELECT_ENDPOINT | endp);
	return inportb(D12_DATA);
}

unsigned char
D12_ReadLastTransactionStatus(unsigned char endp)
{
	outportb(D12_COMMAND, READ_LAST_TRANSACTION_STATUS | endp);
	return inportb(D12_DATA);
}

unsigned char
D12_ReadEndpointStatus(unsigned char endp)
{
	outportb(D12_COMMAND, 0x80 | endp);
	return inportb(D12_DATA);
}

int
D12_ReadEndpoint(unsigned char endp, void* buf, unsigned char maxlen)
{
	unsigned char* dst = (unsigned char*)buf;
	unsigned char i, dummy, len;

	/* ǂݍރGh|CgI܂B
	 * MĂȂ΁A܂ǂݍ߂܂B
	 * * D12_FULLEMPTY=1M(FULL)Ӗ܂B
	 *   EMPTY(D12_FULLEMPTY=0)Ƃ͋tȂ̂ŁAӂĂB
	 */
	if(!(D12_SelectEndpoint(endp) & D12_FULLEMPTY)) return 0;

	/* Mf[^ǂݍ݂܂B */
	outportb(D12_COMMAND, READ_BUFFER);
	dummy = inportb(D12_DATA); /* _~[ */
	len   = inportb(D12_DATA); /* MoCg */
	if(len > maxlen) len = maxlen; /* ǂݍ݃oCg؂l */
	for(i = 0; i < len; i++) dst[i] = inportb(D12_DATA);

	/* Mobt@ɂ܂B */
	outportb(D12_COMMAND, CLEAR_BUFFER);

	return len;
}

int
D12_WriteEndpoint(unsigned char endp, const void* ptr, unsigned char len)
{
	const unsigned char* src = (const unsigned char*)ptr;
	unsigned char i;

#if 0
	/* ރGh|CgI܂B
	 * MĂȂ΁A܂߂܂B
	 * * D12_FULLEMPTY=0M(EMPTY)Ӗ܂B
	 *   FULL(D12_FULLEMPTY=1)Ƃ͋tȂ̂ŁAӂĂB
	 */
	if(D12_SelectEndpoint(endp) & D12_FULLEMPTY) return 0;
#else
	/* C^vg]Ŗ]f[^j邱Ƃ悤Ȃ̂ŁA
	 * EMPTY͂߂܂B
	 * A]f[^cĂƂɎ̃f[^ނƁA
	 * {ɑÕf[^]LẐǂ͖mFłB
	 */
	D12_SelectEndpoint(endp);
#endif

	outportb(D12_COMMAND, WRITE_BUFFER);
	outportb(D12_DATA,   0); /* _~[ */
	outportb(D12_DATA, len); /* MoCg */
	for(i = 0; i < len; i++) outportb(D12_DATA, src[i]);

	/* MJn܂B */
	outportb(D12_COMMAND, VALIDATE_BUFFER);

	return len;
}

void
D12_SetEndpointStatus(unsigned char endp, unsigned char stall)
{
	outportb(D12_COMMAND, SET_ENDPOINT_STATUS | endp);
	outportb(D12_DATA, stall ? 0x01 : 0x00);
}

void
D12_AcknowledgeSetup(unsigned char endp)
{
	outportb(D12_COMMAND, endp);
	outportb(D12_COMMAND, ACKNOWLEDGE_SETUP);
	/* Acknowledge SetuṕAEP0OUT(0)܂EP0IN(1)ɑ΂Ă̂ݔs܂B
	 * EP0OUT(0)ւ̔s̏ꍇAD12_ReadEndpoint()ĂłƎv܂A
	 * D12_ReadEndpoint()̒ōŌɍsClear Buffer͖łB
	 * EP0łSETUPM`Acknowledge SetupԂClear/Validate BufferłB
	 * Acknowledge SetupɂClear Buffers܂B(x͗Lł)
	 */
	if(endp == 0) outportb(D12_COMMAND, CLEAR_BUFFER);
}

void
D12_SendResume()
{
	outportb(D12_COMMAND, SEND_RESUME);
}

unsigned short
D12_ReadCurrentFrameNumber()
{
	unsigned char lo, hi;
	outportb(D12_COMMAND, READ_CURRENT_FRAME_NUMBER);
	lo = inportb(D12_DATA);
	hi = inportb(D12_DATA);
	return hi << 8 | lo;
}

unsigned short
D12_ReadChipID()
{
	unsigned short lo, hi;
	outportb(D12_COMMAND, READ_CHIP_ID);
	lo = inportb(D12_DATA);
	hi = inportb(D12_DATA);
	return hi << 8 | lo;
}
