/*	
 *	d12usr.c
 *
 *	PDIUSBD12 [U[vO
 *	Copyright (c) 1997 Philips Semiconductors
 *	Copyright (c) 2002 Naoyuki Sawa
 *
 *	* Mon Aug 16 12:30:00 JST 2002 Naoyuki Sawa
 *	- 쐬JnB
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <piece.h>
#include <s1c33cpu.h>
#include "c33209.h"
#include "d12ci.h"
#include "d12sys.h"

#define N_SCTS	384
#define SCTSIZ	512
unsigned char scts[N_SCTS][SCTSIZ];

void log(const char* fmt, ...);

void ep2_rxdone(int st);
void ep2_txdone(int st);
void standard_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
void class_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
USBINFO ui = {
	/* USBC mode */
	D12_ENDP_NONISO,	/* mode */
	/* USBC raw event */
	NULL,			/* bus_reset */
	NULL,			/* ep1_rxdone */
	NULL,			/* ep1_txdone */
	ep2_rxdone,		/* ep2_rxdone */
	ep2_txdone,		/* ep2_txdone */
	/* Device request */
	standard_request,	/* standard_request */
	class_request,		/* class_request */
	NULL,			/* vendor_request */
};

/****************************************************************************
 *	fBXNv^
 ****************************************************************************/

static unsigned char DeviceDesc[18] = {
	0x12,			// bLength		
	0x01,			// bDescriptorType	foCXfBXNv^
	0x10,0x01,		// bcdUSB		USB Version 1.1
	0x00,			// bDeviceClass		eC^[tF[Xɐݒ
	0x00,			// bDeviceSubClass	gp
	0x00,			// bDeviceProtocol	gp
	EP0_PACKET_SIZE,	// bMaxPacketSize0	EP0̍őpPbgTCY
	0x19,0x0e,		// idVendor		PCE_VENDOR_ID
	0x12,0x10,		// idProduct		PCE_PRODUCT_ID+0x12
	0x00,0x01,		// bcdDevice		foCX[Xԍ 1.00
	0x00,			// iManufacture		gp
	0x00,			// iProduct		gp
	0x00,			// iSerialNumber	gp
	0x01			// bNumConfiguraions	RtBO[V1
};
static unsigned char ConfigDesc[9] = {
	0x09,			// bLength		
	0x02,			// bDescriptorType	RtBO[VfBXNv^
	0x20,0x00,		// wTotalLength		CFG(9)+I/F(9)+EP(7*2)=32=0x20
	0x01,			// bNumberIntefaces	C^[tF[X1
	0x01,			// bConfigurationValue	RtBO[V1Ȃ̂#1̂
	0x00,			// iConfiguration	gp
	0x80,			// bmAttributes		oXp[foCXA[gEFCNAbv
	0x32			// MaxPower		d 100mA (100/2=50)
};
static unsigned char InterfaceDesc[9] = {
	0x09,			// bLength		
	0x04,			// bDescriptorType	C^[tF[XEfBXNv^
	0x00,			// bInterfaceNumber	C^[tF[X#0
	0x00,			// bAlternateSetting	I^l[gZbg͖
	0x02,			// bNumEndpointes	Gh|Cg2
	0x08,			// bInterfaceClass	Mass storage class
	0x06,			// bInterfaceSubClass	SCSI transparent
	0x50,			// bInterfaceProtocol	Bulk only
	0x00			// iInterface		gp
};
static unsigned char EndpointDesc[2][7] = {
{
	0x07,			// bLength		
	0x05,			// bDescriptorType	Gh|CgfBXNv^
	0x82,			// bEndpointAddress	oNIN
	0x02,			// bAttributes		oN]
	EP2_NONISO_PACKET_SIZE,0x00, // wMaxPacketize	]oCg
	0x00,			// bInterval		C^[o[ms] (gp)
},{
	0x07,			// bLength		
	0x05,			// bDescriptorType	Gh|CgfBXNv^
	0x02,			// bEndpointAddress	oNOUT
	0x02,			// bAttributes		oN]
	EP2_NONISO_PACKET_SIZE,0x00, // wMaxPacketize	]oCg
	0x00,			// bInterval		C^[o[ms] (gp)
}
};

static unsigned char*
dscpy(unsigned char *ptr, const void *src, int len)
{
	memcpy(ptr, src, len);
	return ptr + len;
}

static unsigned char tmpbuff[256]; /* SfBXNv^[ȃTCYKv */

/****************************************************************************
 *	Mass storage (Bulk-Only)
 ****************************************************************************/

/* Command Block Wrapper */
typedef struct _CBW {
	unsigned int dCBWSignature;		/*  0- 3 */
	unsigned int dCBWTag;			/*  4- 7 */
	unsigned int dCBWDataTransferLength;	/*  8-11 */
	unsigned char bmCBWFlags;		/*    12 */
	unsigned char bCBWLUN		:4;	/*    13 */
	unsigned char			:4;
	unsigned char bCBWCBLength	:5;	/*    14 */
	unsigned char			:3;
	unsigned char CBWCB[16];		/* 15-30 */
} CBW;
#define CBW_SIZE	31 /* sizeof(CBW)ł̓pfBÔŃ_! */
#define CBW_SIGNATURE	0x43425355 /* 'USBC' */

/* Command Status Wrapper */
typedef struct _CSW {
	unsigned int dCSWSignature;		/*  0- 3 */
	unsigned int dCSWTag;			/*  4- 7 */
	unsigned int dCSWDataResidue;		/*  8-11 */
	unsigned char bCSWStatus;		/*    12 */
} CSW;
#define CSW_SIZE	13 /* sizeof(CSW)ł̓pfBÔŃ_! */
#define CSW_SIGNATURE	0x53425355 /* 'USBS' */

CBW cbw; /* CBWMp     */
CSW csw; /* CSWM\p */

unsigned char* data_ptr; /* Mobt@|C^(NULL:ACh) */
int data_len; /* MTCY     */
int data_cnt; /* McTCY */

/* INQUERY f[^ */
unsigned char inquiry_data[36] = {
	0x00,	/*  0 */
	0x80,	/*  1 */ /* <-[oufBXNȂ̂0x00ɂƂ낾A0x00ɂWindowsFĂȂ     */
			 /*   s̕iUSBL[ƂӖɃ[oufBXNɂȂĂƂ݂ƁAΏ@̖͗l */
	0x02,	/*  2 */
	0x02,	/*  3 */
	0x1f,	/*  4 */
	0x00,	/*  5 */
	0x00,	/*  6 */
	0x00,	/*  7 */
	'P',	/*  8 */
	'I',	/*  9 */
	'E',	/* 10 */
	'C',	/* 11 */
	'E',	/* 12 */
	'L',	/* 13 */
	'A',	/* 14 */
	'B',	/* 15 */
	'U',	/* 16 */
	'S',	/* 17 */
	'B',	/* 18 */
	'R',	/* 19 */
	'E',	/* 20 */
	'M',	/* 21 */
	'O',	/* 22 */
	'V',	/* 23 */
	'A',	/* 24 */
	'B',	/* 25 */
	'L',	/* 26 */
	'E',	/* 27 */
	'D',	/* 28 */
	'I',	/* 29 */
	'S',	/* 30 */
	'K',	/* 31 */
	'0',	/* 32 */
	'.',	/* 33 */
	'0',	/* 34 */
	'0',	/* 35 */
};

/* READ CAPACITY f[^ */
unsigned char capacity_data[8] = {
	0x00,0x00,(N_SCTS-1)>> 8 & 0xff,(N_SCTS-1)>> 0 & 0xff, /* ZN^ł͂ȂőZN^ԍ */
	0x00,0x00, SCTSIZ   >> 8 & 0xff, SCTSIZ   >> 0 & 0xff,
};

unsigned char mode_sense_data[4 + 13] = {
	/* header */
	4+13-1,	/* Philips̃TvƂȂ-1ĂB̃og1oCg͏OdlȂ̂ȁH */
		/* ȑOPhilips̃Tv͂񂺂TCYԈĂ݂...(4+4-1H)              */
	0x00,
	0x00,
	0x00,
	/* page */
	0x86,
	0x0b,
	0x00,
	SCTSIZ >> 8 & 0xff, /*{{block size*/
	SCTSIZ >> 0 & 0xff, /*}}block size*/
	0x00,               /*{{block count*/ /* Philips̃Tvł͂ȂAll[...gĂȂ񂾂낤H */
	0x00,
	0x00,
	N_SCTS >> 8 & 0xff,
	N_SCTS >> 0 & 0xff, /*}}block count*/
	0xff,
	0x00,
	0x00,
};

/****************************************************************************
 *	WNGXg
 ****************************************************************************/

/* * AUDIO  HID NXł́ÃNGXg͗Ȃ̂ŁAȂĂ܂B
 * * MASS STORAGE NXł ENDPOINT_HALT NÂ߂ɎĝŎĂ݂܂B
 *   ܂mFĂ܂B
 */
static void
clear_feature(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	int endp;

	switch(req->bmRequestType & 0x1f) { /* recipient */
	case 0x00: /* device */
		switch(req->wValue) {
		case 1: /* device remote wakeup */
			break; /* T|[g */
		case 2: /* test mode */
			break; /* T|[g */
		}
		break;
	case 0x01: /* interface */
		/* no features */
		break;
	case 0x02: /* endpoint */
		switch(req->wValue) {
		case 0: /* endpoint halt */
			endp = req->wIndex & 0x7f;
			if(endp > 2) break; /* EPx */
			endp *= 2; /* EP0=0or1/EP1=2or3/EP2=4or5 */
			if(req->wIndex & 0x80) endp += 1; /* EPx IN */
			D12_SetEndpointStatus(endp, 0); /* STALL */
			usb_ep0_tx(NULL, 0);
			/*{{MASS STORAGE BULK-ONLY L̂ӂ܂F */
			/*  BULK IÑXg[CBWorf[^̑MfȂ̂ŁAAɕKfR\CSWMXe[WƂȂ܂*/
			if(endp == 5) {
				D12_WriteEndpoint(5, &csw, CSW_SIZE);
				data_ptr = NULL;
			}
			/*}}*/
			return;
		}
		break;
	}

	usb_ep0_stall();
}

static void
set_address(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	D12_SetAddressEnable(req->wValue, 1);
	usb_ep0_tx(NULL, 0);
}

static void
get_descriptor(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	unsigned char* p;

	switch(req->wValue & 0xff00) {
	case 0x0100: /* Device */
		/* TODO: */
		usb_ep0_tx(DeviceDesc, sizeof DeviceDesc);
		return;
	case 0x0200: /* Configuration */
		/* TODO: */
		p = tmpbuff;
		p = dscpy(p, ConfigDesc, sizeof ConfigDesc);
		p = dscpy(p, InterfaceDesc, sizeof InterfaceDesc);
		p = dscpy(p, EndpointDesc, sizeof EndpointDesc);
		usb_ep0_tx(tmpbuff, p - tmpbuff);
		return;
	case 0x0300: /* String */
	case 0x0400: /* Interface */
	case 0x0500: /* Endpoint */
	case 0x0600: /* Device qualifier */
	case 0x0700: /* Other speed configuration */
	case 0x0800: /* Interface power */
		break; /* Ή */
	}

	usb_ep0_stall();
}

static void
set_configuration(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	D12_SetEndpointEnable(req->wValue != 0);
	usb_ep0_tx(NULL, 0);
}

static void
set_interface(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	usb_ep0_tx(NULL, 0);
}

static void (*standard_request_table[])(const DEVICE_REQUEST* req, const unsigned char* dat, int len) = {
/* 0*/	NULL/*get_status*/,
/* 1*/	clear_feature,
/* 2*/	NULL,
/* 3*/	NULL/*set_feature*/,
/* 4*/	NULL,
/* 5*/	set_address,
/* 6*/	get_descriptor,
/* 7*/	NULL/*set_descriptor*/,
/* 8*/	NULL/*get_configuration*/,
/* 9*/	set_configuration,
/*10*/	NULL/*get_interface*/,
/*11*/	set_interface,
/*12*/	NULL/*synch_frame*/,
};
#define N_STANDARD_REQUEST_TABLE (sizeof standard_request_table / sizeof *standard_request_table)

/****************************************************************************
 *
 ****************************************************************************/

void
standard_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
//log("STD(%d)", req->bRequest);
	if(req->bRequest >= N_STANDARD_REQUEST_TABLE ||
	   standard_request_table[req->bRequest] == NULL) {
		usb_ep0_stall();
		return;
	}
	standard_request_table[req->bRequest](req, dat, len);
}

/* MASS STORAGE BULK-ONLYL̃NXNGXgB0xffn܂邱ƂɒӁI */
void
class_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
//log("CLS(%d)", req->bRequest);
	switch(req->bRequest) {
	case 0xff: /* Bulk-Only Mass Storage Reset */
		/* TODO: tF[YBXg[Ԃ͈ێ܂B */
		usb_ep0_tx(NULL, 0);
		return;
	case 0xfe: /* Get Max LUN */
		/* PLUN̏ꍇSTALLOK(Bulk-Onlydl) */
		break;
	}
	usb_ep0_stall();
}

void
ep2_rxdone(int st)
{
	int op, len, pos;

	/* AChԂȂ... */
	if(data_ptr == NULL) {
		len = D12_ReadEndpoint(4, &cbw, CBW_SIZE);
		op = cbw.CBWCB[0];
		csw.dCSWSignature = CSW_SIGNATURE;
		csw.dCSWTag = cbw.dCBWTag;
		csw.dCSWDataResidue = cbw.dCBWDataTransferLength;
		csw.bCSWStatus = 0x00;
		switch(op) {
		case 0x12: /* INQUIRY */
			data_ptr = inquiry_data;
			data_len = sizeof inquiry_data;
			data_cnt = 0;
			csw.dCSWDataResidue -= data_len;
			ep2_txdone(D12_INT_ENDP2IN);
			return;
		case 0x25: /* READ CAPACITY */
			data_ptr = capacity_data;
			data_len = sizeof capacity_data;
			data_cnt = 0;
			csw.dCSWDataResidue -= data_len;
			ep2_txdone(D12_INT_ENDP2IN);
			return;
		case 0x28: /* READ10 */
			pos = cbw.CBWCB[2] << 24 |
			      cbw.CBWCB[3] << 16 |
			      cbw.CBWCB[4] <<  8 |
			      cbw.CBWCB[5] <<  0;
			data_ptr = &scts[pos][0];
			data_len = cbw.dCBWDataTransferLength;
			data_cnt = 0;
			csw.dCSWDataResidue -= data_len;
			ep2_txdone(D12_INT_ENDP2IN);
			return;
		case 0x1a: /* MODE SENSE */
			if((cbw.CBWCB[2] & 0x3f) == 0x06 ||
			   (cbw.CBWCB[2] & 0x3f) == 0x3f) {
				data_ptr = mode_sense_data;
				//data_len = sizeof mode_sense_data;
				data_len = cbw.dCBWDataTransferLength; /* sSTALLĂApfBO̕y */
				data_cnt = 0;
				//csw.dCSWDataResidue -= data_len;
				csw.dCSWDataResidue -= sizeof mode_sense_data; /* pfBÕTCYʒm̂ */
				ep2_txdone(D12_INT_ENDP2IN);
				return;
			}
			break;
		case 0x00: /* TEST UNIT READY */
		case 0x1e: /* PREVENT ALLOWD REMOVAL */
		case 0x2f: /* VERIFY(10) */
		case 0x1b: /* START STOP UNIT */ /* <-uovI񂾂Ƃɂꂪ܂B */
						 /*    ̃^C~OŃLbVtbVĂƊҁB */
						 /*    ^XNgCUSBfoCX~͂ȂȂ̂ŁAőpB */
			data_ptr = (unsigned char*)0x1000000;
			data_len = 0;
			data_cnt = 0;
			csw.dCSWDataResidue -= data_len;
			ep2_txdone(D12_INT_ENDP2IN);
			return;
		case 0x2a: /* WRITE10 */
			pos = cbw.CBWCB[2] << 24 |
			      cbw.CBWCB[3] << 16 |
			      cbw.CBWCB[4] <<  8 |
			      cbw.CBWCB[5] <<  0;
			data_ptr = &scts[pos][0];
			data_len = cbw.dCBWDataTransferLength;
			data_cnt = 0;
			csw.dCSWDataResidue -= data_len;
			/*ep2_rxdone(D12_INT_ENDP2OUT);*/
			return;
		case 0x23: /* READLONG */
			break;
		default:
			log("CBW(%02x)", op);
			break;
		}
		csw.bCSWStatus = 0x02;
		D12_SetEndpointStatus(4, 1);
		D12_SetEndpointStatus(5, 1);
	/* f[^MȂ... */
	} else {
		/* Mf[^i[B */
		len = D12_ReadEndpoint(4, data_ptr, EP2_NONISO_PACKET_SIZE);
		data_ptr += len;
		data_cnt += len;
		if(data_cnt == data_len) { /* MȂCSWԑ */
			D12_WriteEndpoint(5, &csw, CSW_SIZE);
			data_ptr = NULL;
		}
	}
}

void
ep2_txdone(int st)
{
	int len;

	/* ʓIȃRg[]̃f[^Xe[WAoN]̂悤ɁA
	 * f[^]̍ŌオőpPbgTCYor[TCYpPbgłKv͂Ȃ݂łB
	 * (Philips̃TvΏĂȂŁAzg͕KvȂ̂Ȃ...)
	 * ȂƂmFł́AzXg->foCXւ512KB]ōŌ̃pPbgőTCY҂ŏI邯ǁA
	 * ă[TCYpPbg͑Ă܂B
	 * ƂƂ́AfoCX->zXgł[TCYpPbgőTCYŏI[Kv͂Ȃ񂾂Ǝv܂B
	 */

	/* AChԂȂ... */
	if(data_ptr == NULL) {
		/* ԂCSWMʒmȂ̂ŁA܂ */
	} else {
		/* f[^MACSW𑗐M */
		if(data_cnt == data_len) {
			D12_WriteEndpoint(5, &csw, CSW_SIZE);
			data_ptr = NULL;
		/* f[^M */
		} else {
			len = data_len - data_cnt;
			if(len > EP2_NONISO_PACKET_SIZE) len = EP2_NONISO_PACKET_SIZE;
			D12_WriteEndpoint(5, data_ptr, len);
			data_ptr += len;
			data_cnt += len;
		}
	}
}
